{***************************************************************}
{* Calculation of eigenvalues and eigenvectors of a 2x2 matrix *}
{***************************************************************}

program eigenval;

{$mode objfpc}{$H+}

uses
  SysUtils, Crt, ReadCon;

type
  TMatrix2 = array[0..1, 0..1] of Real;
  TValues  = array of Real;
  TVector  = array of Real;
  TVectors = array of TVector;
  TSValues = array of string;
  TSVectors = array of string;

var
  N, I, J: Integer;
  rA, rB, rC: Real;
  sMess: string;
  bEmpty: Boolean;
  aMatrix2: TMatrix2;
  aValues: TValues;
  aVectors: TVectors;
  aSValues: TSValues;
  aSVectors: TSVectors;

{ Integer power of a real number }

function Power(R: Real; N: Integer): Real;

var
  I: Integer;
  Pow: Real;

begin
  Pow := 1;
  for I := 1 to Abs(N) do
    Pow := Pow * R;
  if N < 0 then
    Pow := 1 / Pow;
  Power := Pow;
end;

{ Format of a real number to specified number of decimal digits }

function RFormat(R: Real; F: Integer): string;

var
  R0: Real;
  SR: string;

begin
  SR := '';
  if R = 0 then
    SR := '0'
  else begin
    if F >= 0 then begin
      R0 := Round(R * Power(10, F)) / Power(10, F);
      if Abs(R0) < Power(10, -F) then
        SR := FloatToStrF(R, ffExponent, F, 0)
      else
        SR := FloatToStr(R0);
    end;
  end;
  RFormat := SR;
end;

{ Read matrix values from the keyboard }

procedure ReadMatrix2(out Matrix2: TMatrix2; out Mess: string);

var
  I, J: Integer;

begin
  Mess := '';
  for I := 0 to 1 do begin
    for J := 0 to 1 do begin
      Write('Matrix element (', I + 1, ', ', J + 1, ')? ');
      ReadReal(Matrix2[I, J], bEmpty, Mess);
      if (Mess = '') and bEmpty then begin
        Mess := 'Missing matrix value';
        Exit;
      end;
    end;
    if Mess <> '' then
      Exit;
  end;
end;

{ Solve 2nd degree equation }

procedure SolveEquation2(A, B, C: Real; var Roots: TValues; out Mess: string);

var
  Delta: Real;

begin
  Roots[0] := 0; Roots[1] := 0; Mess := '';
  Delta := Sqr(B) - 4 * A * C;
  if Delta < 0 then
    Mess := 'no real solution'
  else begin
    Roots[0] := (-B + Sqrt(Delta)) / (2*A);
    Roots[1] := (-B - Sqrt(Delta)) / (2*A);
  end;
end;

begin
  ClrScr;
  Writeln('Eigenvalues and eigenvectors of a 2x2 matrix.');
  Writeln('=============================================');
  Writeln;
  Writeln('Enter matrix elements: ');
  ReadMatrix2(aMatrix2, sMess);
  if sMess = '' then begin
    // Do the calculations
    SetLength(aValues, 2); SetLength(aSValues, 2);
    rA := 1;
    rB := -(aMatrix2[0, 0] + aMatrix2[1, 1]);
    rC := aMatrix2[0, 0] * aMatrix2[1, 1] - aMatrix2[0, 1] * aMatrix2[1, 0];
    SolveEquation2(rA, rB, rC, aValues, sMess);
    if sMess = '' then begin
      // Real eigenvalues
      if aValues[0] = aValues[1] then
        N := 1
      else
        N := 2;
      SetLength(aVectors, N); SetLength(aSVectors, N);
      for I := 0 to N - 1 do begin
        SetLength(aVectors[I], 2);
        aVectors[I][0] := 1;
        if aMatrix2[0, 1] = 0 then
          aVectors[I][1] := 0
        else
          aVectors[I][1] := (aValues[I] - aMatrix2[0, 0]) / aMatrix2[0, 1];
      end;
    end
    else begin
      // Complex eigenvalues
      aSValues[0] := 'not real ';
      aSValues[1] := 'not real ';
    end;
    // Display calculation results
    for I := 0 to N - 1 do begin
      aSValues[I] := RFormat(aValues[I], 3);
      aSVectors[I] := '( ';
      for J := 0 to Length(aVectors[I]) - 1 do begin
        if J > 0 then
          aSVectors[I] := aSVectors[I] + ' ; ';
        aSVectors[I] := aSVectors[I] + RFormat(aVectors[I][J], 3);
      end;
      aSVectors[I] := aSVectors[I] + ' )';
    end;
    Writeln;
    if N = 1 then begin
      Writeln('Eigenvalue:  ', aSValues[0]);
      Writeln('Eigenvector: ', aSVectors[0]);
    end
    else begin
      Writeln('Eigenvalues:  ', aSValues[0], ', ', aSValues[1]);
      Writeln('Eigenvectors: ', aSVectors[0], ', ', aSVectors[1]);
    end;
  end
  else begin
    // Invalid user input (missing values)
    Writeln;
    Writeln('Data error:', sMess);
  end;
end.

